/*
 * This file is part of SionTower.
 *
 * 
 * David Saltares M&aacute;rquez (C) 2011
 * <david.saltares@gmail.com>
 *
 * 
 * SionTower examples are free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License ad
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) ant later version.
 *
 * SionTower examples are distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SionTower examples.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SIONTOWER_TRUNK_SRC_INCLUDE_SOUND3D_H_
#define SIONTOWER_TRUNK_SRC_INCLUDE_SOUND3D_H_

#include <vector>

#include <OGRE/Ogre.h>

#include "soundFX.h"

//! Clase para simular sonidos en el espacio con una fuente y una entidad que escucha

/**
 *  @author David Saltares M&aacute;rquez
 *  @date 29-12-2010
 *
 *  Esta clase utiliza el recurso SoundFX o el gestor SoundFXManager para
 *  reproducir efectos de sonido con un efecto pseudo-3D. Sound3D asocia un
 *  nodo emisor de sonido, un nodo que escucha y un efecto a reproducir. El
 *  volumen de la reproducci&oacute;n var&iacute;a en funci&oacute;n de la distancia y se produce un
 *  efecto est&eacute;reo en los altavoces en funci&oacute;n del &aacute;ngulo entre los nodos.
 *
 *  Es necesario actualizar el Sound3D en cada iteraci&oacute;n del bucle de juego
 *  para que el volumen y el efecto est&eacute;reo reflejen lo que ocurre en la escena.
 *
 *  Ejemplo:
 *
 *  \code
 *  Sound3D grito("disparo.wav");
 *  grito.setListenerNode(nodoPersonaje);
 *  grito.setSourceNode(nodoEnemigo);
 *
 *  ...
 *
 *  while(!salir) {
 *      ...
 *
 *      if(enemigoHerido)
 *          grito.play();
 *      ...
 *
 *      Sound3D::update3DSound();
 *
 *      ...
 *  }
 *
 *  \endcode
 */
class Sound3D {
    public:
        /**
         *  Constructor
         *
         *  @param name nombre del recurso tipo SoundFX a cargar.
         *  @param group nombre del grupo al que pertenece el recurso. Por
         *  defecto es el grupo "General".
         *  @param listenerNode puntero al nodo que escucha el sonido.
         *  @param sourceNode puntero al nodo que emite el sonido.
         */
        Sound3D(const Ogre::String& name,
                const Ogre::String& group = Ogre::ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME,
                Ogre::SceneNode* listenerNode = 0,
                Ogre::SceneNode* sourceNode = 0);

        /**
         *  Constructor
         *
         *  @param s puntero inteligente al sonido asociado al efecto 3D.
         *  @param listenerNode puntero al nodo que escucha el sonido.
         *  @param sourceNode puntero al nodo que emite el sonido.
         */
        Sound3D(SoundFXPtr s,
                Ogre::SceneNode* listenerNode = 0,
                Ogre::SceneNode* sourceNode = 0);

        /**
         *  Destructor
         */
        ~Sound3D();

        /**
         *  @param loop n&uacute;mero de veces que se reproduce el efecto de sonido.
         *  -1 equivale a infinitas repeticiones, 0 equivale a 1 reproducci&oacute;n.
         *  En general, si bucle = n, el efecto se reproducir&aacute; n + 1 veces.
         *
         *  Reproduce el efecto 3D.
         */
        void play(int loop = 0);

        /**
         *  Actualiza el volumen y el efecto est&eacute;reo de todos los efectos 3D en
         *  reproducci&oacute;n. Es necesario llamar a este m&eacute;todo en cada iteraci&oacute;n
         *  del bucle principal.
         */
        static void update3DSound();

        /**
         *  @param channel canal que acaba de finalizar su reproducci&oacute;n.
         *
         *  M&eacute;todo que utiliza el callback para gestionar internamente el fin
         *  de la reproducci&oacute;n de un efecto en un canal.
         */
        static void channelFinished(int channel);

        /**
         *  @return puntero al nodo que escucha el efecto.
         */
        const Ogre::SceneNode* getListenerNode() const;

        /**
         *  @return puntero al nodo que emite el efecto.
         */
        const Ogre::SceneNode* getSourceNode() const;

        /**
         *  @param e puntero al nuevo nodo que escucha el efecto.
         */
        void setListenerNode(const Ogre::SceneNode* listenerNode);

        /**
         *  @param f puntero al nuevo nodo que emite el efecto.
         */
        void setSourceNode(const Ogre::SceneNode* sourceNode);

    private:
        SoundFXPtr _soundFXPtr;
        const Ogre::SceneNode* _listenerNode;
        const Ogre::SceneNode* _sourceNode;
        
        // Vector con los sonidos 3D en reproducci&oacute;n junto a su canal asociado.
        static std::vector<std::pair<int, Sound3D*> > _sounds;
        // Vector con todos los canales de forma que canales[n] indica la
        // posici&oacute;n en sonidos de ese canal.
        static std::vector<int> _channels;
        static int _numUsedChannels;

        void update(int channel);
};

/**
 *  @param channel canal que ha terminado de reproducir un efecto.
 *
 *  Funci&oacute;n de callback para SDL_Mixer que nos avisa cuando un canal ha
 *  terminado de reproducir un efecto.
 */
void channelFinished(int channel);
    


#endif  // SIONTOWER_TRUNK_SRC_INCLUDE_SOUND3D_H_
